@@ -66,35 +66,20 @@ def get_Kstat():
             name, unused, value = kstat.split()
             Kstat[namespace + name] = D(value)
 
-    Kstats = [
-        "hw.pagesize",
-        "hw.physmem",
-        "kern.maxusers",
-        "vm.kmem_map_free",
-        "vm.kmem_map_size",
-        "vm.kmem_size",
-        "vm.kmem_size_max",
-        "vm.kmem_size_min",
-        "vm.kmem_size_scale",
-        "vm.stats",
-        "vm.swap_total",
-        "vm.swap_reserved",
-        "kstat.zfs",
-        "vfs.zfs"
-    ]
     Kstat = {}
     load_proc_kstats('/proc/spl/kstat/zfs/arcstats',
-            'kstat.zfs.misc.arcstats.')
+                     'kstat.zfs.misc.arcstats.')
     load_proc_kstats('/proc/spl/kstat/zfs/zfetchstats',
-            'kstat.zfs.misc.zfetchstats.')
+                     'kstat.zfs.misc.zfetchstats.')
     load_proc_kstats('/proc/spl/kstat/zfs/vdev_cache_stats',
-            'kstat.zfs.misc.vdev_cache_stats.')
+                     'kstat.zfs.misc.vdev_cache_stats.')
 
     return Kstat
 
+
 def div1():
     sys.stdout.write("\n")
-    for i in xrange(18):
+    for i in range(18):
         sys.stdout.write("%s" % "----")
     sys.stdout.write("\n")
 
@@ -188,28 +173,27 @@ def get_arc_summary(Kstat):
 
     output['memory_throttle_count'] = fHits(memory_throttle_count)
 
-    ### ARC Misc. ###
+    # ARC Misc.
     deleted = Kstat["kstat.zfs.misc.arcstats.deleted"]
     mutex_miss = Kstat["kstat.zfs.misc.arcstats.mutex_miss"]
-    recycle_miss = Kstat["kstat.zfs.misc.arcstats.recycle_miss"]
 
-    ### ARC Misc. ###
+    # ARC Misc.
     output["arc_misc"] = {}
     output["arc_misc"]["deleted"] = fHits(deleted)
-    output["arc_misc"]['recycle_miss'] = fHits(recycle_miss)
     output["arc_misc"]['mutex_miss'] = fHits(mutex_miss)
     output["arc_misc"]['evict_skips'] = fHits(mutex_miss)
 
-    ### ARC Sizing ###
+    # ARC Sizing
     arc_size = Kstat["kstat.zfs.misc.arcstats.size"]
-    mru_size = Kstat["kstat.zfs.misc.arcstats.p"]
+    mru_size = Kstat["kstat.zfs.misc.arcstats.mru_size"]
+    mfu_size = Kstat["kstat.zfs.misc.arcstats.mfu_size"]
     target_max_size = Kstat["kstat.zfs.misc.arcstats.c_max"]
     target_min_size = Kstat["kstat.zfs.misc.arcstats.c_min"]
     target_size = Kstat["kstat.zfs.misc.arcstats.c"]
 
     target_size_ratio = (target_max_size / target_min_size)
 
-    ### ARC Sizing ###
+    # ARC Sizing
     output['arc_sizing'] = {}
     output['arc_sizing']['arc_size'] = {
         'per': fPerc(arc_size, target_max_size),
@@ -228,7 +212,7 @@ def get_arc_summary(Kstat):
         'num': fBytes(target_size),
     }
 
-    ### ARC Hash Breakdown ###
+    # ARC Hash Breakdown
     output['arc_hash_break'] = {}
     output['arc_hash_break']['hash_chain_max'] = Kstat[
         "kstat.zfs.misc.arcstats.hash_chain_max"
@@ -247,29 +231,16 @@ def get_arc_summary(Kstat):
         ]
 
     output['arc_size_break'] = {}
-    if arc_size > target_size:
-        mfu_size = (arc_size - mru_size)
-        output['arc_size_break']['recently_used_cache_size'] = {
-            'per': fPerc(mru_size, arc_size),
-            'num': fBytes(mru_size),
-        }
-        output['arc_size_break']['frequently_used_cache_size'] = {
-            'per': fPerc(mfu_size, arc_size),
-            'num': fBytes(mfu_size),
-        }
+    output['arc_size_break']['recently_used_cache_size'] = {
+        'per': fPerc(mru_size, mru_size + mfu_size),
+        'num': fBytes(mru_size),
+    }
+    output['arc_size_break']['frequently_used_cache_size'] = {
+        'per': fPerc(mfu_size, mru_size + mfu_size),
+        'num': fBytes(mfu_size),
+    }
 
-    elif arc_size < target_size:
-        mfu_size = (target_size - mru_size)
-        output['arc_size_break']['recently_used_cache_size'] = {
-            'per': fPerc(mru_size, target_size),
-            'num': fBytes(mru_size),
-        }
-        output['arc_size_break']['frequently_used_cache_size'] = {
-            'per': fPerc(mfu_size, target_size),
-            'num': fBytes(mfu_size),
-        }
-
-    ### ARC Hash Breakdown ###
+    # ARC Hash Breakdown
     hash_chain_max = Kstat["kstat.zfs.misc.arcstats.hash_chain_max"]
     hash_chains = Kstat["kstat.zfs.misc.arcstats.hash_chains"]
     hash_collisions = Kstat["kstat.zfs.misc.arcstats.hash_collisions"]
@@ -290,27 +261,25 @@ def get_arc_summary(Kstat):
 
 
 def _arc_summary(Kstat):
-    ### ARC Sizing ###
+    # ARC Sizing
     arc = get_arc_summary(Kstat)
 
     sys.stdout.write("ARC Summary: (%s)\n" % arc['health'])
 
     sys.stdout.write("\tMemory Throttle Count:\t\t\t%s\n" %
-            arc['memory_throttle_count'])
+                     arc['memory_throttle_count'])
     sys.stdout.write("\n")
 
-    ### ARC Misc. ###
+    # ARC Misc.
     sys.stdout.write("ARC Misc:\n")
     sys.stdout.write("\tDeleted:\t\t\t\t%s\n" % arc['arc_misc']['deleted'])
-    sys.stdout.write("\tRecycle Misses:\t\t\t\t%s\n" %
-            arc['arc_misc']['recycle_miss'])
     sys.stdout.write("\tMutex Misses:\t\t\t\t%s\n" %
-            arc['arc_misc']['mutex_miss'])
+                     arc['arc_misc']['mutex_miss'])
     sys.stdout.write("\tEvict Skips:\t\t\t\t%s\n" %
-            arc['arc_misc']['mutex_miss'])
+                     arc['arc_misc']['mutex_miss'])
     sys.stdout.write("\n")
 
-    ### ARC Sizing ###
+    # ARC Sizing
     sys.stdout.write("ARC Size:\t\t\t\t%s\t%s\n" % (
         arc['arc_sizing']['arc_size']['per'],
         arc['arc_sizing']['arc_size']['num']
@@ -348,21 +317,21 @@ def _arc_summary(Kstat):
 
     sys.stdout.write("\n")
 
-    ### ARC Hash Breakdown ###
+    # ARC Hash Breakdown
     sys.stdout.write("ARC Hash Breakdown:\n")
     sys.stdout.write("\tElements Max:\t\t\t\t%s\n" %
-            arc['arc_hash_break']['elements_max'])
+                     arc['arc_hash_break']['elements_max'])
     sys.stdout.write("\tElements Current:\t\t%s\t%s\n" % (
         arc['arc_hash_break']['elements_current']['per'],
         arc['arc_hash_break']['elements_current']['num'],
         )
     )
     sys.stdout.write("\tCollisions:\t\t\t\t%s\n" %
-            arc['arc_hash_break']['collisions'])
+                     arc['arc_hash_break']['collisions'])
     sys.stdout.write("\tChain Max:\t\t\t\t%s\n" %
-            arc['arc_hash_break']['chain_max'])
+                     arc['arc_hash_break']['chain_max'])
     sys.stdout.write("\tChains:\t\t\t\t\t%s\n" %
-            arc['arc_hash_break']['chains'])
+                     arc['arc_hash_break']['chains'])
 
 
 def get_arc_efficiency(Kstat):
@@ -492,7 +461,7 @@ def _arc_efficiency(Kstat):
     arc = get_arc_efficiency(Kstat)
 
     sys.stdout.write("ARC Total accesses:\t\t\t\t\t%s\n" %
-            arc['total_accesses'])
+                     arc['total_accesses'])
     sys.stdout.write("\tCache Hit Ratio:\t\t%s\t%s\n" % (
         arc['cache_hit_ratio']['per'],
         arc['cache_hit_ratio']['num'],
@@ -703,7 +672,7 @@ def _l2arc_summary(Kstat):
         else:
             sys.stdout.write("(HEALTHY)\n")
         sys.stdout.write("\tLow Memory Aborts:\t\t\t%s\n" %
-                arc['low_memory_aborts'])
+                         arc['low_memory_aborts'])
         sys.stdout.write("\tFree on Write:\t\t\t\t%s\n" % arc['free_on_write'])
         sys.stdout.write("\tR/W Clashes:\t\t\t\t%s\n" % arc['rw_clashes'])
         sys.stdout.write("\tBad Checksums:\t\t\t\t%s\n" % arc['bad_checksums'])
@@ -711,7 +680,7 @@ def _l2arc_summary(Kstat):
         sys.stdout.write("\n")
 
         sys.stdout.write("L2 ARC Size: (Adaptive)\t\t\t\t%s\n" %
-                arc["l2_arc_size"]["adative"])
+                         arc["l2_arc_size"]["adative"])
         sys.stdout.write("\tCompressed:\t\t\t%s\t%s\n" % (
             arc["l2_arc_size"]["actual"]["per"],
             arc["l2_arc_size"]["actual"]["num"],
@@ -724,17 +693,17 @@ def _l2arc_summary(Kstat):
         )
         sys.stdout.write("\n")
 
-        if arc["l2_arc_evicts"]['lock_retries'] + \
-                arc["l2_arc_evicts"]["reading"] > 0:
+        if arc["l2_arc_evicts"]['lock_retries'] != '0' or \
+           arc["l2_arc_evicts"]["reading"] != '0':
             sys.stdout.write("L2 ARC Evicts:\n")
             sys.stdout.write("\tLock Retries:\t\t\t\t%s\n" %
-                    arc["l2_arc_evicts"]['lock_retries'])
+                             arc["l2_arc_evicts"]['lock_retries'])
             sys.stdout.write("\tUpon Reading:\t\t\t\t%s\n" %
-                    arc["l2_arc_evicts"]["reading"])
+                             arc["l2_arc_evicts"]["reading"])
             sys.stdout.write("\n")
 
         sys.stdout.write("L2 ARC Breakdown:\t\t\t\t%s\n" %
-                arc['l2_arc_breakdown']['value'])
+                         arc['l2_arc_breakdown']['value'])
         sys.stdout.write("\tHit Ratio:\t\t\t%s\t%s\n" % (
             arc['l2_arc_breakdown']['hit_ratio']['per'],
             arc['l2_arc_breakdown']['hit_ratio']['num'],
@@ -748,7 +717,7 @@ def _l2arc_summary(Kstat):
         )
 
         sys.stdout.write("\tFeeds:\t\t\t\t\t%s\n" %
-                arc['l2_arc_breakdown']['feeds'])
+                         arc['l2_arc_breakdown']['feeds'])
         sys.stdout.write("\n")
 
         sys.stdout.write("L2 ARC Writes:\n")
@@ -779,39 +748,13 @@ def _l2arc_summary(Kstat):
 def get_dmu_summary(Kstat):
     output = {}
 
-    zfetch_bogus_streams = Kstat["kstat.zfs.misc.zfetchstats.bogus_streams"]
-    zfetch_colinear_hits = Kstat["kstat.zfs.misc.zfetchstats.colinear_hits"]
-    zfetch_colinear_misses = \
-            Kstat["kstat.zfs.misc.zfetchstats.colinear_misses"]
     zfetch_hits = Kstat["kstat.zfs.misc.zfetchstats.hits"]
     zfetch_misses = Kstat["kstat.zfs.misc.zfetchstats.misses"]
-    zfetch_reclaim_failures = \
-            Kstat["kstat.zfs.misc.zfetchstats.reclaim_failures"]
-    zfetch_reclaim_successes = \
-            Kstat["kstat.zfs.misc.zfetchstats.reclaim_successes"]
-    zfetch_streams_noresets = \
-            Kstat["kstat.zfs.misc.zfetchstats.streams_noresets"]
-    zfetch_streams_resets = Kstat["kstat.zfs.misc.zfetchstats.streams_resets"]
-    zfetch_stride_hits = Kstat["kstat.zfs.misc.zfetchstats.stride_hits"]
-    zfetch_stride_misses = Kstat["kstat.zfs.misc.zfetchstats.stride_misses"]
 
     zfetch_access_total = (zfetch_hits + zfetch_misses)
-    zfetch_colinear_total = (zfetch_colinear_hits + zfetch_colinear_misses)
-    zfetch_health_count = (zfetch_bogus_streams)
-    zfetch_reclaim_total = (zfetch_reclaim_successes + zfetch_reclaim_failures)
-    zfetch_streams_total = (zfetch_streams_resets + zfetch_streams_noresets +
-            zfetch_bogus_streams)
-    zfetch_stride_total = (zfetch_stride_hits + zfetch_stride_misses)
     output['zfetch_access_total'] = zfetch_access_total
 
     if zfetch_access_total > 0:
-
-        output['file_level_prefetch'] = {}
-        if zfetch_health_count > 0:
-            output['file_level_prefetch']['health'] = 'DEGRADED'
-        else:
-            output['file_level_prefetch']['health'] = 'HEALTHY'
-
         output['dmu'] = {}
         output['dmu']['efficiency'] = {}
         output['dmu']['efficiency']['value'] = fHits(zfetch_access_total)
@@ -824,57 +767,6 @@ def get_dmu_summary(Kstat):
             'num': fHits(zfetch_misses),
         }
 
-        output['dmu']['colinear'] = {}
-        output['dmu']['colinear']['value'] = fHits(zfetch_colinear_total)
-        output['dmu']['colinear']['hit_ratio'] = {
-            'per': fPerc(zfetch_colinear_hits, zfetch_colinear_total),
-            'num': fHits(zfetch_colinear_hits),
-        }
-        output['dmu']['colinear']['miss_ratio'] = {
-            'per': fPerc(zfetch_colinear_misses, zfetch_colinear_total),
-            'num': fHits(zfetch_colinear_misses),
-        }
-
-        output['dmu']['stride'] = {}
-        output['dmu']['stride']['value'] = fHits(zfetch_stride_total)
-        output['dmu']['stride']['hit_ratio'] = {
-            'per': fPerc(zfetch_stride_hits, zfetch_stride_total),
-            'num': fHits(zfetch_stride_hits),
-        }
-        output['dmu']['stride']['miss_ratio'] = {
-            'per': fPerc(zfetch_stride_misses, zfetch_stride_total),
-            'num': fHits(zfetch_stride_misses),
-        }
-
-        output['dmu_misc'] = {}
-        if zfetch_health_count > 0:
-            output['dmu_misc']['status'] = "FAULTED"
-        else:
-            output['dmu_misc']['status'] = ""
-
-        output['dmu_misc']['reclaim'] = {}
-        output['dmu_misc']['reclaim']['value'] = fHits(zfetch_reclaim_total)
-        output['dmu_misc']['reclaim']['successes'] = {
-            'per': fPerc(zfetch_reclaim_successes, zfetch_reclaim_total),
-            'num': fHits(zfetch_reclaim_successes),
-        }
-        output['dmu_misc']['reclaim']['failure'] = {
-            'per': fPerc(zfetch_reclaim_failures, zfetch_reclaim_total),
-            'num': fHits(zfetch_reclaim_failures),
-        }
-
-        output['dmu_misc']['streams'] = {}
-        output['dmu_misc']['streams']['value'] = fHits(zfetch_streams_total)
-        output['dmu_misc']['streams']['plus_resets'] = {
-            'per': fPerc(zfetch_streams_resets, zfetch_streams_total),
-            'num': fHits(zfetch_streams_resets),
-        }
-        output['dmu_misc']['streams']['neg_resets'] = {
-            'per': fPerc(zfetch_streams_noresets, zfetch_streams_total),
-            'num': fHits(zfetch_streams_noresets),
-        }
-        output['dmu_misc']['streams']['bogus'] = fHits(zfetch_bogus_streams)
-
     return output
 
 
@@ -883,12 +775,8 @@ def _dmu_summary(Kstat):
     arc = get_dmu_summary(Kstat)
 
     if arc['zfetch_access_total'] > 0:
-        sys.stdout.write("File-Level Prefetch: (%s)" %
-                arc['file_level_prefetch']['health'])
-        sys.stdout.write("\n")
-
-        sys.stdout.write("DMU Efficiency:\t\t\t\t\t%s\n" %
-                arc['dmu']['efficiency']['value'])
+        sys.stdout.write("DMU Prefetch Efficiency:\t\t\t\t\t%s\n" %
+                         arc['dmu']['efficiency']['value'])
         sys.stdout.write("\tHit Ratio:\t\t\t%s\t%s\n" % (
             arc['dmu']['efficiency']['hit_ratio']['per'],
             arc['dmu']['efficiency']['hit_ratio']['num'],
@@ -902,80 +790,16 @@ def _dmu_summary(Kstat):
 
         sys.stdout.write("\n")
 
-        sys.stdout.write("\tColinear:\t\t\t\t%s\n" %
-                arc['dmu']['colinear']['value'])
-        sys.stdout.write("\t  Hit Ratio:\t\t\t%s\t%s\n" % (
-            arc['dmu']['colinear']['hit_ratio']['per'],
-            arc['dmu']['colinear']['hit_ratio']['num'],
-            )
-        )
-
-        sys.stdout.write("\t  Miss Ratio:\t\t\t%s\t%s\n" % (
-            arc['dmu']['colinear']['miss_ratio']['per'],
-            arc['dmu']['colinear']['miss_ratio']['num'],
-            )
-        )
-
-        sys.stdout.write("\n")
-
-        sys.stdout.write("\tStride:\t\t\t\t\t%s\n" %
-                arc['dmu']['stride']['value'])
-        sys.stdout.write("\t  Hit Ratio:\t\t\t%s\t%s\n" % (
-            arc['dmu']['stride']['hit_ratio']['per'],
-            arc['dmu']['stride']['hit_ratio']['num'],
-            )
-        )
-
-        sys.stdout.write("\t  Miss Ratio:\t\t\t%s\t%s\n" % (
-            arc['dmu']['stride']['miss_ratio']['per'],
-            arc['dmu']['stride']['miss_ratio']['num'],
-            )
-        )
-
-        sys.stdout.write("\n")
-        sys.stdout.write("DMU Misc: %s\n" % arc['dmu_misc']['status'])
-
-        sys.stdout.write("\tReclaim:\t\t\t\t%s\n" %
-                arc['dmu_misc']['reclaim']['value'])
-        sys.stdout.write("\t  Successes:\t\t\t%s\t%s\n" % (
-            arc['dmu_misc']['reclaim']['successes']['per'],
-            arc['dmu_misc']['reclaim']['successes']['num'],
-            )
-        )
-
-        sys.stdout.write("\t  Failures:\t\t\t%s\t%s\n" % (
-            arc['dmu_misc']['reclaim']['failure']['per'],
-            arc['dmu_misc']['reclaim']['failure']['num'],
-            )
-        )
-
-        sys.stdout.write("\n\tStreams:\t\t\t\t%s\n" %
-                arc['dmu_misc']['streams']['value'])
-        sys.stdout.write("\t  +Resets:\t\t\t%s\t%s\n" % (
-            arc['dmu_misc']['streams']['plus_resets']['per'],
-            arc['dmu_misc']['streams']['plus_resets']['num'],
-            )
-        )
-
-        sys.stdout.write("\t  -Resets:\t\t\t%s\t%s\n" % (
-            arc['dmu_misc']['streams']['neg_resets']['per'],
-            arc['dmu_misc']['streams']['neg_resets']['num'],
-            )
-        )
-
-        sys.stdout.write("\t  Bogus:\t\t\t\t%s\n" %
-                arc['dmu_misc']['streams']['bogus'])
-
 
 def get_vdev_summary(Kstat):
     output = {}
 
     vdev_cache_delegations = \
-            Kstat["kstat.zfs.misc.vdev_cache_stats.delegations"]
+        Kstat["kstat.zfs.misc.vdev_cache_stats.delegations"]
     vdev_cache_misses = Kstat["kstat.zfs.misc.vdev_cache_stats.misses"]
     vdev_cache_hits = Kstat["kstat.zfs.misc.vdev_cache_stats.hits"]
     vdev_cache_total = (vdev_cache_misses + vdev_cache_hits +
-            vdev_cache_delegations)
+                        vdev_cache_delegations)
 
     output['vdev_cache_total'] = vdev_cache_total
 
@@ -1024,7 +848,8 @@ def _tunable_summary(Kstat):
 
     values = {}
     for name in names:
-        with open("/sys/module/zfs/parameters/" + name) as f: value = f.read()
+        with open("/sys/module/zfs/parameters/" + name) as f:
+            value = f.read()
         values[name] = value.strip()
 
     descriptions = {}
@@ -1033,7 +858,7 @@ def _tunable_summary(Kstat):
         try:
             command = ["/sbin/modinfo", "zfs", "-0"]
             p = Popen(command, stdin=PIPE, stdout=PIPE,
-                    stderr=PIPE, shell=False, close_fds=True)
+                      stderr=PIPE, shell=False, close_fds=True)
             p.wait()
 
             description_list = p.communicate()[0].strip().split('\0')
@@ -1048,11 +873,11 @@ def _tunable_summary(Kstat):
                         descriptions[name] = description
             else:
                 sys.stderr.write("%s: '%s' exited with code %i\n" %
-                        (sys.argv[0], command[0], p.returncode))
+                                 (sys.argv[0], command[0], p.returncode))
                 sys.stderr.write("Tunable descriptions will be disabled.\n")
         except OSError as e:
             sys.stderr.write("%s: Cannot run '%s': %s\n" %
-                    (sys.argv[0], command[0], e.strerror))
+                             (sys.argv[0], command[0], e.strerror))
             sys.stderr.write("Tunable descriptions will be disabled.\n")
 
     sys.stdout.write("ZFS Tunable:\n")
@@ -1064,7 +889,7 @@ def _tunable_summary(Kstat):
         if alternate_tunable_layout:
             format = "\t%s=%s\n"
 
-        if show_tunable_descriptions and descriptions.has_key(name):
+        if show_tunable_descriptions and name in descriptions:
             sys.stdout.write("\t# %s\n" % descriptions[name])
 
         sys.stdout.write(format % (name, values[name]))
@@ -1091,21 +916,23 @@ def zfs_header():
 def usage():
     sys.stdout.write("Usage: arc_summary.py [-h] [-a] [-d] [-p PAGE]\n\n")
     sys.stdout.write("\t -h, --help           : "
-            "Print this help message and exit\n")
+                     "Print this help message and exit\n")
     sys.stdout.write("\t -a, --alternate      : "
-            "Show an alternate sysctl layout\n")
+                     "Show an alternate sysctl layout\n")
     sys.stdout.write("\t -d, --description    : "
-            "Show the sysctl descriptions\n")
+                     "Show the sysctl descriptions\n")
     sys.stdout.write("\t -p PAGE, --page=PAGE : "
-            "Select a single output page to display,\n")
+                     "Select a single output page to display,\n")
     sys.stdout.write("\t                        "
-            "should be an integer between 1 and " + str(len(unSub)) + "\n\n")
+                     "should be an integer between 1 and " +
+                     str(len(unSub)) + "\n\n")
     sys.stdout.write("Examples:\n")
     sys.stdout.write("\tarc_summary.py -a\n")
     sys.stdout.write("\tarc_summary.py -p 4\n")
     sys.stdout.write("\tarc_summary.py -ad\n")
     sys.stdout.write("\tarc_summary.py --page=2\n")
 
+
 def main():
     global show_tunable_descriptions
     global alternate_tunable_layout
@@ -1136,9 +963,9 @@ def main():
     if 'p' in args:
         try:
             pages.append(unSub[int(args['p']) - 1])
-        except IndexError , e:
+        except IndexError:
             sys.stderr.write('the argument to -p must be between 1 and ' +
-                    str(len(unSub)) + '\n')
+                             str(len(unSub)) + '\n')
             sys.exit()
     else:
         pages = unSub
@@ -1148,5 +975,6 @@ def main():
         page(Kstat)
         div2()
 
+
 if __name__ == '__main__':
     main()
